mod accel;
mod config;
#[allow(dead_code)]
mod command;
mod modem;
mod serial;
mod types;
mod gps;

use anyhow;
use std::{thread::{self, JoinHandle}, time::Duration};
use esp_idf_hal::peripherals::Peripherals;
use esp_idf_hal::prelude::*;
use esp_idf_hal::serial::{Pins, config::Config, Serial, UART1, Uart};
use embedded_hal::digital::v2::OutputPin;

use types::*;

fn main() -> anyhow::Result<()> {
    esp_idf_sys::link_patches();

    let dp = Peripherals::take().expect("error taking peripherals");

    println!("Rust main thread: {:?}", thread::current());

    let mut threads: Vec<JoinHandle<anyhow::Result<_>>> = vec![];

    // // Rx/Tx pins for the GPS modem
    let gps_rx = dp.pins.gpio32;
    let gps_tx = dp.pins.gpio33;
    // // UART interface for the GPS modem
    let gps_uart = dp.uart2;


    let (gps_sender, receiver) = std::sync::mpsc::sync_channel::<Msg>(3);
    //let accel_sender = gps_sender.clone();

    //let _ = gps::main(gps_tx, gps_rx, gps_uart, gps_sender)?;
    //threads.push(thread::spawn(move || accel::main(accel_sender)));

    // ==================================
    // MODEM INITIALIZATION AND MAIN LOOP
    // ==================================

    // LilyGo TTGO T-Call sim800l board serial pins.
    let modem_rx = dp.pins.gpio26;
    let modem_tx = dp.pins.gpio27;
    // LilyGo TTGO T-Call sim800l board power / reset pins.
    let modem_pwrkey = dp.pins.gpio4.into_output()?;
    let modem_rst = dp.pins.gpio5.into_output()?;
    let modem_power = dp.pins.gpio23.into_output()?;
    // UART interface for the GSM modem
    let modem_uart = dp.uart1;

    let serial_pins = Pins {
        tx: modem_tx,
        rx: modem_rx,
        cts: None,
        rts: None,
    };

    let serial: Serial<UART1, _, _> = Serial::new(
        modem_uart,
        serial_pins,
        Config::default().baudrate(Hertz(115200)),
    )?;

    let (tx, rx) = serial.split();
    type PwrkeyOutput = esp_idf_hal::gpio::Gpio4<esp_idf_hal::gpio::Output>;
    type ResetOutput = esp_idf_hal::gpio::Gpio5<esp_idf_hal::gpio::Output>;
    type PowerOutput = esp_idf_hal::gpio::Gpio23<esp_idf_hal::gpio::Output>;

    let mut mdm: modem::Modem<UART1, PwrkeyOutput, ResetOutput, PowerOutput> = modem::Modem::new(tx, rx, modem_pwrkey, modem_rst, modem_power, receiver);

    let mqtt_username = include_str!("../secret/username").trim();
    let mqtt_password = include_str!("../secret/password").trim();

    threads.push(thread::spawn(move || gps::main(gps_tx, gps_rx, gps_uart, gps_sender.clone())));

    println!("======================= MAIN =======================");
    mdm.init().unwrap_or(());
    let _ = mdm.echo(false).unwrap_or(());
    println!("resetting modem ... ");
    println!("======================= MODEM =======================");
    let _ = mdm.mqtt_send_position_loop("51.158.66.64", 7887, mqtt_username, mqtt_password).unwrap_or(());
    let _ = mdm.tcp_close_connection().unwrap_or(());
    thread::sleep(Duration::from_millis(1500));
    panic!("rebooting");
}
